@@ -35,7 +35,15 @@ class Test_KDE_PDF(object):
         classes = [lambda xx, bw: sp.stats.gaussian_kde(xx, bw_method=bw),
                    lambda xx, bw: kale.KDE(xx, bandwidth=bw, kernel=kernel)]
         for mm in methods:
-            kde_list = [cc(aa, mm).pdf(grid) for cc in classes]
+            kde_list = []
+            for cc in classes:
+                try:
+                    test = cc(aa, mm).density(grid, probability=True)[1]
+                except AttributeError:
+                    test = cc(aa, mm).pdf(grid)
+
+                kde_list.append(test)
+
             print("method: {}".format(mm))
             print("\t" + utils.stats_str(kde_list[0]))
             print("\t" + utils.stats_str(kde_list[1]))
@@ -55,7 +63,7 @@ class Test_KDE_PDF(object):
 
         data = [aa, bb]
         edges = [utils.spacing(dd, 'lin', 30, stretch=0.5) for dd in data]
-        cents = [utils.midpoints(ee, 'lin') for ee in edges]
+        cents = [utils.midpoints(ee, log=False) for ee in edges]
 
         xe, ye = np.meshgrid(*edges, indexing='ij')
         xc, yc = np.meshgrid(*cents, indexing='ij')
@@ -66,8 +74,15 @@ class Test_KDE_PDF(object):
         classes = [lambda xx, bw: sp.stats.gaussian_kde(xx, bw_method=bw),
                    lambda xx, bw: kale.KDE(xx, bandwidth=bw, kernel=kernel)]
         for mm in methods:
-            kdes_list = [cc(data, mm).pdf(grid).reshape(xc.shape).T
-                         for cc in classes]
+            kdes_list = []
+            for cc in classes:
+                try:
+                    test = cc(data, mm).density(grid, probability=True)[1].reshape(xc.shape).T
+                except AttributeError:
+                    test = cc(data, mm).pdf(grid).reshape(xc.shape).T
+
+                kdes_list.append(test)
+
             assert_true(np.allclose(kdes_list[0], kdes_list[1]))
 
         return
@@ -81,13 +96,13 @@ class Test_KDE_PDF(object):
         aa = np.random.uniform(*EXTR, NUM)
 
         egrid = utils.spacing(aa, 'lin', 2000, stretch=0.5)
-        cgrid = utils.midpoints(egrid, 'lin')
+        cgrid = utils.midpoints(egrid, log=False)
         delta = np.diff(egrid)
 
         boundaries = [None, EXTR]
         for bnd in boundaries:
             kde = kale.KDE(aa, kernel=kernel)
-            pdf = kde.pdf(cgrid, reflect=bnd)
+            pdf = kde.density(cgrid, reflect=bnd, probability=True)[1]
 
             # If the kernel's support is infinite, then all points outside of boundaries should be
             # nonzero; if it's finite-supported, then only some of them (near edges) will be
@@ -129,9 +144,9 @@ class Test_KDE_PDF(object):
         yy = np.concatenate([yy, np.random.choice(yy, NUM-yy.size)])
 
         data = [xx, yy]
-        edges = [kale.utils.spacing(aa, 'lin', 30) for aa in [xx, yy]]
-        egrid = [kale.utils.spacing(ee, 'lin', 100, stretch=0.5) for ee in edges]
-        cgrid = [kale.utils.midpoints(ee, 'lin') for ee in egrid]
+        edges = [utils.spacing(aa, 'lin', 30) for aa in [xx, yy]]
+        egrid = [utils.spacing(ee, 'lin', 100, stretch=0.5) for ee in edges]
+        cgrid = [utils.midpoints(ee, log=False) for ee in egrid]
         width = [np.diff(ee) for ee in egrid]
 
         xc, yc = np.meshgrid(*cgrid, indexing='ij')
@@ -150,7 +165,7 @@ class Test_KDE_PDF(object):
             None
         ]
         for jj, reflect in enumerate(reflections):
-            pdf_1d = kde.pdf(grid, reflect=reflect)
+            pdf_1d = kde.density(grid, reflect=reflect, probability=True)[1]
             pdf = pdf_1d.reshape(hist.shape)
 
             inside = np.ones_like(pdf_1d, dtype=bool)
@@ -207,8 +222,8 @@ class Test_KDE_PDF(object):
 
         kde = kale.KDE(data, bandwidth=bandwidth, kernel=kernel)
 
-        edges = [kale.utils.spacing(dd, 'lin', 200, stretch=0.1) for dd in data]
-        cents = [kale.utils.midpoints(ee, 'lin') for ee in edges]
+        edges = [utils.spacing(dd, 'lin', 200, stretch=0.1) for dd in data]
+        cents = [utils.midpoints(ee, log=False) for ee in edges]
         widths = [np.diff(ee) for ee in edges]
         # area = widths[0][:, np.newaxis] * widths[1][np.newaxis, :]
 
@@ -220,10 +235,12 @@ class Test_KDE_PDF(object):
 
         for par in range(2):
             xx = cents[par]
-            pdf_2d = kde.pdf(xx, params=par)
+            pdf_2d = kde.density(xx, params=par, probability=True)[1]
             kde_1d = kale.KDE(data[par, :], bandwidth=bandwidth, kernel=kernel)
-            pdf_1d = kde_1d.pdf(xx)
+            pdf_1d = kde_1d.density(xx, probability=True)[1]
             # print("matrix : ", kde.bandwidth.matrix, kde_1d.bandwidth.matrix)
+            print(f"pdf_1d = {utils.stats_str(pdf_1d)}")
+            print(f"pdf_2d = {utils.stats_str(pdf_2d)}")
             assert_true(np.allclose(pdf_2d, pdf_1d, rtol=1e-3))
 
             for pdf, ls, lw in zip([pdf_2d, pdf_1d], ['-', '--'], [1.5, 3.0]):
@@ -270,17 +287,17 @@ class Test_KDE_PDF_Box(Test_KDE_PDF):
 
     def test_reflect_1d(self):
         print("\n|Test_KDE_PDF:test_reflect_1d()|")
-        self.reflect_1d(kale.kernels.Box_Asym)
+        self.reflect_1d(kale.kernels.Box)
         return
 
     def test_reflect_2d(self):
         print("\n|Test_KDE_PDF:test_reflect_2d()|")
-        self.reflect_2d(kale.kernels.Box_Asym)
+        self.reflect_2d(kale.kernels.Box)
         return
 
     def test_pdf_params_fixed_bandwidth(self):
         print("\n|Test_KDE_PDF_Box:test_pdf_params_fixed_bandwidth()|")
-        self.pdf_params_fixed_bandwidth(kale.kernels.Box_Asym)
+        self.pdf_params_fixed_bandwidth(kale.kernels.Box)
         return
 
 
@@ -302,7 +319,7 @@ class Test_KDE_Resample(object):
 
         data = [aa, bb]
         edges = [utils.spacing(dd, 'lin', 100, stretch=1.0) for dd in data]
-        cents = [utils.midpoints(ee, 'lin') for ee in edges]
+        cents = [utils.midpoints(ee, log=False) for ee in edges]
 
         xe, ye = np.meshgrid(*edges, indexing='ij')
         xc, yc = np.meshgrid(*cents, indexing='ij')
@@ -318,7 +335,8 @@ class Test_KDE_Resample(object):
             # Make sure the two distributions resemble eachother
             ks, pv = sp.stats.ks_2samp(samp_1d, samp_2d)
             # Calibrated to the above seed-value of `9235`
-            assert_true(pv > 0.5)
+            print("{}, pv = {}".format(ii, pv))
+            assert_true(pv > 0.05)
 
         return
 
@@ -343,7 +361,10 @@ class Test_KDE_Resample(object):
         # Add an array of uniform values at location `ii`, make sure they are preserved in resample
         for ii in range(3):
             test = np.array(data)
-            test = np.insert(test, ii, norm*np.ones_like(test[0]), axis=0)
+            tt = norm * np.ones_like(test[0])
+            idx = np.random.choice(tt.size, tt.size//2)
+            tt[idx] *= -1
+            test = np.insert(test, ii, tt, axis=0)
 
             # Construct KDE
             kde3d = kale.KDE(test)
@@ -352,7 +373,7 @@ class Test_KDE_Resample(object):
             samples = kde3d.resample(NUM, keep=ii)
             # Make sure the uniform values are still the same
             param_samp = samples[ii]
-            assert_true(np.allclose(param_samp, norm))
+            assert_true(np.all(np.isclose(param_samp, norm) | np.isclose(param_samp, -norm)))
 
             # Make sure the other two parameters are consistent (KS-test) with input data
             samples = np.delete(samples, ii, axis=0)
@@ -360,10 +381,10 @@ class Test_KDE_Resample(object):
                 stuff = [samples[jj], data[jj]]
                 ks, pv = sp.stats.ks_2samp(*stuff)
                 msg = "{} {} :: {:.2e} {:.2e}".format(ii, jj, ks, pv)
-                print("\t" + kale.utils.stats_str(stuff[0]))
-                print("\t" + kale.utils.stats_str(stuff[1]))
+                print("\t" + utils.stats_str(stuff[0]))
+                print("\t" + utils.stats_str(stuff[1]))
                 print(msg)
-                assert_true(pv > 0.01)
+                assert_true(pv > 0.05)
 
         return
 
@@ -415,8 +436,8 @@ class Test_KDE_Resample(object):
                 ks, pv = sp.stats.ks_2samp(*stuff)
                 msg = "\t:: {:.2e} {:.2e}".format(ks, pv)
                 print(msg)
-                print("\t\t" + kale.utils.stats_str(stuff[0]))
-                print("\t\t" + kale.utils.stats_str(stuff[1]))
+                print("\t\t" + utils.stats_str(stuff[0]))
+                print("\t\t" + utils.stats_str(stuff[1]))
                 assert_true(pv > 0.01)
 
         return
@@ -475,9 +496,9 @@ class Test_KDE_Resample(object):
         yy = np.concatenate([yy, np.random.choice(yy, NUM-yy.size)])
 
         data = [xx, yy]
-        edges = [kale.utils.spacing(aa, 'lin', 30) for aa in [xx, yy]]
-        egrid = [kale.utils.spacing(ee, 'lin', 100, stretch=0.5) for ee in edges]
-        cgrid = [kale.utils.midpoints(ee, 'lin') for ee in egrid]
+        edges = [utils.spacing(aa, 'lin', 30) for aa in [xx, yy]]
+        egrid = [utils.spacing(ee, 'lin', 100, stretch=0.5) for ee in edges]
+        cgrid = [utils.midpoints(ee, log=False) for ee in egrid]
         # width = [np.diff(ee) for ee in egrid]
 
         xc, yc = np.meshgrid(*cgrid, indexing='ij')
@@ -527,6 +548,54 @@ class Test_KDE_Resample(object):
         return
 
 
+class Test_KDE_Construct_From_Hist(object):
+
+    @classmethod
+    def setup_class(cls):
+        np.random.seed(9865)
+
+    def test_compare_1d(self):
+        print("\n|Test_KDE_Construct_From_Hist:test_compare_1d()|")
+
+        # Create a pdf and histogram
+        xx = np.linspace(-5, 5, 10001)
+        pdf = np.exp(-xx*xx/2) / np.sqrt(2*np.pi)
+        bins = np.zeros(xx.size + 1)
+        dx = xx[1] - xx[0]
+        bins[:-1] = xx - 0.5 * dx
+        bins[-1] = xx[-1] + 0.5 * dx
+
+        # Construct a KDE from the histogram
+        kde = kale.KDE.from_hist(bins, pdf)
+
+        # Check that the KDE pdf matches the true pdf
+        xx, pdf_kde = kde.density(xx, probability=True)
+        np.testing.assert_allclose( pdf, pdf_kde, atol=1e-5 )
+
+    def test_compare_2d(self):
+        print("\n|Test_KDE_Construct_From_Hist:test_compare_2d()|")
+
+        # Create a pdf and histogram
+        xx = np.linspace(-5, 5, 101)
+        yy = np.linspace(-5, 5, 101)
+        xx_grid, yy_grid = np.meshgrid( xx, yy )
+        hist = np.exp(-xx_grid*xx_grid/2 - yy_grid*yy_grid/2 - xx_grid*yy_grid/2)
+        bins = np.zeros(xx.size + 1)
+        dx = xx[1] - xx[0]
+        bins[:-1] = xx - 0.5 * dx
+        bins[-1] = xx[-1] + 0.5 * dx
+        bins = np.array([ bins, ]*2)
+
+        # Construct a KDE from the histogram
+        kde = kale.KDE.from_hist(bins, hist)
+
+        # Check that the KDE pdf matches the true pdf
+        points = [ xx_grid.flatten(), yy_grid.flatten() ]
+        points, pdf_kde = kde.density( points, probability=True)
+        pdf = hist / ( hist.sum() * dx * dx )
+        np.testing.assert_allclose( pdf.flatten(), pdf_kde, atol=0.01 )
+
+
 # Run all methods as if with `nosetests ...`
 if __name__ == "__main__":
     run_module_suite()
