@@ -12,7 +12,7 @@ from numpy.testing import run_module_suite
 from nose import tools
 
 import kalepy as kale
-from kalepy import utils
+from kalepy import utils, kernels
 
 
 class Test_Kernels_Generic(object):
@@ -34,9 +34,9 @@ class Test_Kernels_Generic(object):
 
         hh = 1.0
         edges = np.linspace(-10*hh, 10*hh, 10000)
-        cents = kale.utils.midpoints(edges, 'lin')
+        cents = kale.utils.midpoints(edges, log=False)
         # width = np.diff(edges)
-        yy = kernel.evaluate(cents)
+        yy = kernel.evaluate(cents[np.newaxis, :], 1).squeeze()
         # Make sure kernel is callable
         # tools.assert_true(np.allclose(yy, kernel().evaluate(cents)))
 
@@ -94,18 +94,18 @@ class Test_Kernels_Generic(object):
 
     @classmethod
     def _test_evaluate_nd(cls, kernel):
-        print("\n|Test_Kernels_Generic:_test_evaluate_nd()|")
-        print(kernel)
+        # print("\n|Test_Kernels_Generic:_test_evaluate_nd()|")
+        # print(kernel)
 
-        kernels = kale.kernels.get_all_distribution_classes()
+        # kernels = kale.kernels.DISTRIBUTIONS.values()
 
         num_dims = [1, 2, 3, 4]
 
-        for kern in kernels:
-            print("\nkern: ", kern)
-            for ndim in num_dims:
-                print("\nndim: ", ndim)
-                cls.kernel_at_dim(kern, ndim)
+        # for kern in kernels:
+        # print("kernel: ", kernel)
+        for ndim in num_dims:
+            # print("\tndim: ", ndim)
+            cls.kernel_at_dim(kernel, ndim)
 
         return
 
@@ -147,7 +147,7 @@ class Test_Kernels_Generic(object):
 def test_kernels_evaluate():
     print("\n|test_kernels.py:test_kernels_evaluate()|")
 
-    for kernel in kale.kernels.get_all_distribution_classes():
+    for kernel in kale.kernels.DISTRIBUTIONS.values():
         print("Testing '{}'".format(kernel))
         Test_Kernels_Generic._test_evaluate(kernel)
 
@@ -155,25 +155,121 @@ def test_kernels_evaluate():
 
 
 def test_kernels_evaluate_nd():
-    print("\n|test_kernels.py:test_kernels_evaluate_nd()|")
-
-    for kernel in kale.kernels.get_all_distribution_classes():
-        print("Testing '{}'".format(kernel))
+    # print("\n|test_kernels.py:test_kernels_evaluate_nd()|")
+    for kernel in kale.kernels.DISTRIBUTIONS.values():
+        # print("Testing '{}'".format(kernel))
         Test_Kernels_Generic._test_evaluate_nd(kernel)
 
     return
 
 
-'''
-def test_kernels_resample():
-    print("\n|test_kernels.py:test_kernels_resample()|")
+def _test_check_reflect(ndim, weights_flag):
+    reflects_good = [
+        None,
+        [-1.0, 1.0],
+        [-1.0, None],
+        [None, 1.0],
+        [None, None],
+    ]
 
-    for kernel in kale.kernels.get_all_distribution_classes():
-        print("Testing '{}'".format(kernel))
-        Test_Kernels_Generic._test_resample(kernel)
+    reflects_bad = [
+        [None],
+        10.0,
+        [12.0],
+        [None, 1.0, 2.0],
+        [2.0, 1.0],
+    ]
+
+    ndata = 7
+
+    # Generate data
+    data = []
+    for ii in range(ndim):
+        data.append(np.random.uniform(-1.0, 1.0, ndata))
+
+    weights = np.random.uniform(0.0, 1.0, ndata) if weights_flag else None
+    nref = len(reflects_good)
+    shape = [nref] * ndim
+    for inds in np.ndindex(*shape):
+        reflect = [reflects_good[ii] for ii in inds]
+        print("---------")
+        print("inds = {}, reflect = {}".format(inds, reflect))
+        # Make sure good values do not raise errors, and return the appropriate object
+        ref = kernels._check_reflect(reflect, data, weights=weights)
+        if len(ref) != ndim:
+            err = "Returned reflect has length {}, but ndim = {}!".format(len(ref), ndim)
+            raise ValueError(err)
+
+        # Make sure bad values raise errors
+        for jj, bad in enumerate(reflects_bad):
+            jj = jj % ndim
+            reflect[jj] = bad
+            tools.assert_raises(ValueError, kernels._check_reflect, *(reflect, data, weights))
 
     return
-'''
+
+
+def test_check_reflect():
+
+    for ndim in range(1, 4):
+        for wf in [False, True]:
+            _test_check_reflect(ndim, wf)
+
+    return
+
+
+def test_check_reflect_boolean():
+    np.random.seed(12456)
+    ndim = 3
+
+    data = np.random.uniform(-1.0, 1.0, (ndim, 100))
+    data[1] *= 2.0
+
+    ref = kernels._check_reflect(None, data)
+    tools.assert_true(ref is None)
+
+    ref = kernels._check_reflect(False, data)
+    tools.assert_true(ref is None)
+
+    extrema = [utils.minmax(dd) for dd in data]
+    extrema = np.asarray(extrema)
+    extrema[:, 0] *= (1 - kale._NUM_PAD)
+    extrema[:, 1] *= (1 + kale._NUM_PAD)
+
+    # If `True` is Given, values should be set to extrema
+    reflect = kernels._check_reflect(True, data)
+    tools.assert_true(np.allclose(reflect, extrema, atol=1e-10))
+
+    for ii in range(ndim):
+        # True given for a single dimension
+        reflect = [None for ii in range(ndim)]
+        reflect[ii] = True
+        ref = kernels._check_reflect(reflect, data)
+
+        tools.assert_true(np.allclose(ref[ii], extrema[ii], atol=1e-10))
+        tools.assert_true(np.all([ref[jj] is None for jj in range(ndim) if jj != ii]))
+
+        # True given as lower or upper bound for dimension
+        reflect = [None for ii in range(ndim)]
+        # val = np.random.uniform(-10.0, 10.0)
+        val = 12.34
+
+        # Fixed upper-value (must be larger than lower-value!)
+        reflect[ii] = [True, val]
+        ref = kernels._check_reflect(reflect, data)
+        tools.assert_true(np.isclose(ref[ii][0], extrema[ii][0], atol=1e-10))
+        tools.assert_true(ref[ii][1] == val)
+        tools.assert_true(np.all([ref[jj] is None for jj in range(ndim) if jj != ii]))
+
+        # Fixed lower-value (must be less than upper-value!)
+        reflect[ii] = [-val, True]
+        ref = kernels._check_reflect(reflect, data)
+        tools.assert_true(np.isclose(ref[ii][1], extrema[ii][1], atol=1e-10))
+        tools.assert_true(ref[ii][0] == -val)
+        tools.assert_true(np.all([ref[jj] is None for jj in range(ndim) if jj != ii]))
+
+    return
+
 
 # Run all methods as if with `nosetests ...`
 if __name__ == "__main__":
