@@ -13,7 +13,8 @@ from nose.tools import assert_true
 from kalepy import utils
 
 
-class Test_Bound_Indices(utils.Test_Base):
+# class Test_Bound_Indices(utils.Test_Base):
+class Test_Bound_Indices:
 
     def test_1d(self):
         aa = np.random.uniform(*[-100, 100], 1000)
@@ -43,6 +44,140 @@ class Test_Bound_Indices(utils.Test_Base):
         return
 
 
+class Test_Histogram(object):
+
+    @classmethod
+    def setup_class(cls):
+        np.random.seed(9865)
+
+        num_points = 20
+
+        cls.bins = [
+            13,
+            np.linspace(-1.0, 1.0, 11),
+        ]
+
+        cls.data = [
+            np.ones(num_points),
+            np.random.uniform(-1, 1, num_points),
+            np.random.poisson(size=num_points) / np.sqrt(num_points),
+        ]
+
+        cls.weights = [
+            None,
+            np.ones(num_points),
+            np.ones(num_points) / num_points,
+            np.random.uniform(0.0, 10.0, num_points),
+        ]
+        return
+
+    def test_hist_dens_prob(self):
+        for weights in self.weights:
+            self._test_hist_dens_prob(weights)
+
+        return
+
+    def _test_hist_dens_prob(self, weights):
+        for data in self.data:
+            for bins in self.bins:
+                hh, ee = utils.histogram(data, bins,
+                                         weights=weights, density=True, probability=True)
+                hh_true, ee_true = np.histogram(data, bins, weights=weights, density=True)
+
+                if not np.all(ee == ee_true):
+                    print("edges     = ", ee)
+                    print("    truth = ", ee_true)
+                    raise ValueError("Edges do not match!")
+
+                bads = ~np.isclose(hh, hh_true)
+                if np.any(bads):
+                    print("hist      = ", hh)
+                    print("    truth = ", hh_true)
+                    raise ValueError("Histograms do not match!")
+
+        return
+
+    def test_hist(self):
+        for weights in self.weights:
+            self._test_hist(weights)
+
+        return
+
+    def _test_hist(self, weights):
+        for data in self.data:
+            for bins in self.bins:
+                hh, ee = utils.histogram(data, bins,
+                                         weights=weights, density=False, probability=False)
+                hh_true, ee_true = np.histogram(data, bins, weights=weights, density=False)
+
+                if not np.all(ee == ee_true):
+                    print("edges     = ", ee)
+                    print("    truth = ", ee_true)
+                    raise ValueError("Edges do not match!")
+
+                bads = ~np.isclose(hh, hh_true)
+                if np.any(bads):
+                    print("hist      = ", hh)
+                    print("    truth = ", hh_true)
+                    raise ValueError("Histograms do not match!")
+
+        return
+
+    def test_hist_dens(self):
+        for weights in self.weights:
+            self._test_hist_dens(weights)
+
+        return
+
+    def _test_hist_dens(self, weights):
+        for data in self.data:
+            for bins in self.bins:
+                hh, ee = utils.histogram(data, bins,
+                                         weights=weights, density=True, probability=False)
+                hh_true, ee_true = np.histogram(data, bins, weights=weights, density=False)
+                hh_true = hh_true.astype(float) / np.diff(ee_true)
+
+                if not np.all(ee == ee_true):
+                    print("edges     = ", ee)
+                    print("    truth = ", ee_true)
+                    raise ValueError("Edges do not match!")
+
+                bads = ~np.isclose(hh, hh_true)
+                if np.any(bads):
+                    print("hist      = ", hh)
+                    print("    truth = ", hh_true)
+                    raise ValueError("Histograms do not match!")
+
+        return
+
+    def test_hist_prob(self):
+        for weights in self.weights:
+            self._test_hist_prob(weights)
+
+        return
+
+    def _test_hist_prob(self, weights):
+        for data in self.data:
+            for bins in self.bins:
+                hh, ee = utils.histogram(data, bins,
+                                         weights=weights, density=False, probability=True)
+                hh_true, ee_true = np.histogram(data, bins, weights=weights, density=False)
+                hh_true = hh_true.astype(float) / hh_true.sum()
+
+                if not np.all(ee == ee_true):
+                    print("edges     = ", ee)
+                    print("    truth = ", ee_true)
+                    raise ValueError("Edges do not match!")
+
+                bads = ~np.isclose(hh, hh_true)
+                if np.any(bads):
+                    print("hist      = ", hh)
+                    print("    truth = ", hh_true)
+                    raise ValueError("Histograms do not match!")
+
+        return
+
+
 class Test_Midpoints(object):
 
     @classmethod
@@ -57,20 +192,20 @@ class Test_Midpoints(object):
         shp = (12, 14, 16)
         test = np.ones(shp)
         for ii in range(test.ndim):
-            vals = utils.midpoints(test, 'lin', axis=ii)
+            vals = utils.midpoints(test, log=False, axis=ii)
             new_shape = np.array(shp)
             new_shape[ii] -= 1
             assert_true(np.all(vals.shape == new_shape))
             assert_true(np.all(vals == 1.0))
 
-            vals = utils.midpoints(test, 'log', axis=ii)
+            vals = utils.midpoints(test, log=True, axis=ii)
             new_shape = np.array(shp)
             new_shape[ii] -= 1
             assert_true(np.all(vals.shape == new_shape))
             assert_true(np.all(vals == 1.0))
 
         test = np.arange(10)
-        vals = utils.midpoints(test, 'lin')
+        vals = utils.midpoints(test, log=False)
         true = 0.5 * (test[:-1] + test[1:])
         assert_true(np.allclose(vals, true))
         return
@@ -91,14 +226,14 @@ class Test_Midpoints(object):
         ]
 
         for ii, tr in enumerate(truth):
-            vals = utils.midpoints(test, 'lin', axis=ii)
+            vals = utils.midpoints(test, log=False, axis=ii, squeeze=True)
             assert_true(np.all(np.shape(tr) == np.shape(vals)))
             assert_true(np.all(tr == vals))
 
         shp = (4, 5)
         test = np.random.uniform(-1.0, 1.0, np.product(shp)).reshape(shp)
         for ii in range(2):
-            vals = utils.midpoints(test, 'lin', axis=ii)
+            vals = utils.midpoints(test, log=False, axis=ii)
 
             temp = np.moveaxis(test, ii, 0)
             true = temp[:-1, :] + 0.5*np.diff(temp, axis=0)
@@ -126,7 +261,7 @@ class Test_Midpoints(object):
         ]
 
         for ii, tr in enumerate(truth):
-            vals = utils.midpoints(test, 'log', axis=ii)
+            vals = utils.midpoints(test, log=True, axis=ii, squeeze=True)
             assert_true(np.all(np.shape(tr) == np.shape(vals)))
             assert_true(np.allclose(tr, vals))
 
@@ -135,8 +270,8 @@ class Test_Midpoints(object):
         test_lin = 10**test_log
         for ii in range(2):
             # Make sure `midpoints` gives consistent results itself
-            vals_log = utils.midpoints(test_log, 'lin', axis=ii)
-            vals_lin = utils.midpoints(test_lin, 'log', axis=ii)
+            vals_log = utils.midpoints(test_log, log=False, axis=ii)
+            vals_lin = utils.midpoints(test_lin, log=True, axis=ii)
             assert_true(np.all(np.shape(vals_log) == np.shape(vals_lin)))
             assert_true(np.allclose(10**vals_log, vals_lin))
 
@@ -261,7 +396,8 @@ class Test_Trapz(object):
         return
 
 
-class Test_Trapz_Dens_To_Mass(utils.Test_Base):
+# class Test_Trapz_Dens_To_Mass(utils.Test_Base):
+class Test_Trapz_Dens_To_Mass:
 
     def _test_ndim(self, ndim):
         from kalepy import utils
@@ -385,7 +521,7 @@ class Test_Trapz_Dens_To_Mass(utils.Test_Base):
                 temp = wid[tuple(cut)]
                 widths.append(temp)
 
-            wids = np.product(widths, axis=0)
+            wids = np.product(np.array(widths, dtype=object), axis=0).astype(float)
 
             pdf = np.ones_like(grid[0]) * norm
             pmf = utils.trapz_dens_to_mass(pdf, edges, axis=axis)
@@ -415,7 +551,8 @@ class Test_Trapz_Dens_To_Mass(utils.Test_Base):
         return
 
 
-class Test_Cumsum(utils.Test_Base):
+# class Test_Cumsum(utils.Test_Base):
+class Test_Cumsum:
 
     def _brute_force_cumsum(self, vals):
         """Brute-force cumulative sum calculation as comparison
@@ -531,55 +668,11 @@ class Test_Cumsum(utils.Test_Base):
         return
 
 
-class Test_Pre_Pad_Zero(utils.Test_Base):
-
-    def _compare_io(self, inn, out, axis):
-        innsh = np.shape(inn)
-        outsh = np.shape(out)
-        ndim = len(innsh)
-        if axis is None:
-            axis = np.arange(ndim)
-        axis = np.atleast_1d(axis)
-        cut_base = [slice(None) for ii in range(ndim)]
-        for aa in axis:
-            si = innsh[aa]
-            so = outsh[aa]
-            msg = "Output along axis={} is {{fail:}}the correct shape".format(aa)
-            utils.assert_true(so == si+1, msg=msg)
-            cut = [0 if ii == aa else cb for ii, cb in enumerate(cut_base)]
-            msg = "Padding for axis={} is {{fail:}}all zero".format(aa)
-            utils.alltrue(out[tuple(cut)] == 0.0, msg)
-
-        return
-
-    def _test_ndim_axis(self, ndim, axis):
-
-        # Construct a random shape in `ndim` dimensions
-        shape = np.random.randint(2, 7, ndim)
-        # Fill with random values
-        vals = np.random.uniform(-20.0, 20.0, shape)
-
-        res = utils._pre_pad_zero(vals, axis=axis)
-        if axis is None:
-            chk = np.pad(vals, [1, 0])
-            msg = "Output ndim={} without axis does {{fail:}}match numpy result".format(ndim)
-            utils.allclose(res, chk, msg=msg)
-
-        self._compare_io(vals, res, axis)
-        return
-
-    def test(self):
-        for nd in range(1, 5):
-            for ax in [None] + list(range(nd)):
-                self._test_ndim_axis(nd, ax)
-
-        return
-
-
-class Test_Really1D(utils.Test_Base):
+# class Test_Really1D(utils.Test_Base):
+class Test_Really1D:
 
     def _test_vals(self, vals, truth):
-        print("`vv` should be: {} :: shape = {} :: '{}'".format(truth, np.shape(vals), vals))
+        print("`vv` should be: {} :: shape = {} :: '{}'".format(truth, utils.jshape(vals), vals))
         assert_true(utils.really1d(vals) == truth)
         return
 
