@@ -1,26 +1,21 @@
-"""Simple utility methods.
+"""kalepy's internal, utility functions.
 """
 import logging
-import os
-import re
+import copy
 
 import numpy as np
 import scipy as sp
 import scipy.linalg  # noqa
 
 
-'''
-__all__ = [
-    'add_cov', 'array_str', 'ave_std', 'bins', 'check_path', 'cumsum', 'cumtrapz', 'midpoints',
-    'minmax', 'modify_exists', 'percentiles', 'rem_cov',
-    'spacing', 'stats_str',
-    'trapz_nd', 'trapz_dens_to_mass'
-]
-'''
-
-
 def add_cov(data, cov):
-    color_mat = sp.linalg.cholesky(cov)
+    try:
+        color_mat = sp.linalg.cholesky(cov)
+    except Exception:
+        logging.error("Cholesky decomposition failed!")
+        logging.error("cov matrix: {}".format(cov))
+        raise
+
     color_data = np.dot(color_mat.T, data)
     return color_data
 
@@ -48,17 +43,6 @@ def array_str(data, num=3, format=':.2e'):
     return rv
 
 
-def assert_true(val, msg=None):
-    msg_succ, msg_fail = _prep_msg(msg)
-    if not val:
-        raise AssertionError(msg_fail)
-
-    if msg_succ is not None:
-        print(msg_succ)
-
-    return
-
-
 def allclose(xx, yy, msg=None, **kwargs):
     msg_succ, msg_fail = _prep_msg(msg)
     xx = np.atleast_1d(xx)
@@ -95,17 +79,6 @@ def alltrue(xx, msg=None):
     return
 
 
-def ave_std(values, weights=None, **kwargs):
-    """
-    Return the weighted average and (biased[1]) standard deviation.
-
-    [1]: i.e. we are dividing by the size `n` of values, not `n-1`.
-    """
-    average = np.average(values, weights=weights, **kwargs)
-    variance = np.average((values - average)**2, weights=weights, **kwargs)
-    return average, np.sqrt(variance)
-
-
 def bins(*args, **kwargs):
     """Calculate `np.linspace(*args)` and return also centers and widths.
 
@@ -148,69 +121,24 @@ def bound_indices(data, bounds, outside=False):
     return idx
 
 
-# NOTE: this is SLOWER
-def _bound_indices(data, bounds, outside=False):
-    """Find the indices of the `data` array that are bounded by the given `bounds`.
+def cov_keep_vars(matrix, keep, reflect=None):
+    matrix = np.array(matrix)
+    if (keep is None) or (keep is False):
+        return matrix
 
-    If `outside` is True, then indices for values *outside* of the bounds are returned.
-    """
-    data = np.atleast_2d(data)
-    bounds = np.atleast_2d(bounds)
-    ndim, nvals = np.shape(data)
-    # shape = (ndim, 2, nvals)
-    shape = (ndim, nvals)
+    if keep is True:
+        keep = np.arange(matrix.shape[0])
 
-    if outside:
-        idx = np.zeros(shape, dtype=int)
-    else:
-        idx = np.ones(shape, dtype=int)
+    keep = np.atleast_1d(keep)
+    for pp in keep:
+        matrix[pp, :] = 0.0
+        matrix[:, pp] = 0.0
+        # Make sure this isn't also a reflection axis
+        if (reflect is not None) and (reflect[pp] is not None):
+            err = "Cannot both 'keep' and 'reflect' about dimension '{}'".format(pp)
+            raise ValueError(err)
 
-    for ii, bnd in enumerate(bounds):
-        if ((len(bnd) == 1) and (bnd[0] is None)):
-            continue
-
-        if bnd[0] is None:
-            bnd[0] = -np.inf
-        if bnd[1] is None:
-            bnd[1] = +np.inf
-
-        temp = np.searchsorted(bnd, data[ii, :])
-        temp = (temp + outside) % 2
-        idx[ii, :] = temp
-
-    idx = np.product(idx, axis=0).astype(bool)
-
-    return idx
-
-
-def check_path(fname):
-    """Make sure the given path exists. Create directories as needed.
-    """
-    path, fname = os.path.split(fname)
-    if len(path) > 0 and not os.path.exists(path):
-        os.makedirs(path)
-    return
-
-
-def cov_from_var_cor(var, corr):
-    var = np.atleast_1d(var)
-    assert np.ndim(var) == 1, "`var` should be 1D!"
-    ndim = len(var)
-    # Covariance matrix diagonals should be the variance (of each parameter)
-    cov = np.identity(ndim) * var
-
-    if np.isscalar(corr):
-        corr = corr * np.ones((ndim, ndim))
-    elif np.shape(corr) != (ndim, ndim):
-        raise ValueError("`corr` must be either a scalar or (D,D) matrix!")
-
-    # Set the off-diagonals to be the correlation, times the product of standard-deviations
-    for ii, jj in np.ndindex(cov.shape):
-        if ii == jj:
-            continue
-        cov[ii, jj] = np.sqrt(var[ii]) * np.sqrt(var[jj]) * corr[ii, jj]
-
-    return cov
+    return matrix
 
 
 def cumsum(vals, axis=None):
@@ -233,7 +161,7 @@ def cumsum(vals, axis=None):
     """
 
     vals = np.asarray(vals)
-    nd = np.ndim(vals)
+    nd = _ndim(vals)
     if (axis is not None) or (nd == 1):
         return np.cumsum(vals, axis=axis)
 
@@ -264,9 +192,9 @@ def cumtrapz(pdf, edges, prepend=True, axis=None):
     cdf : ndarray of scalar
         Values integrated over the desired axes.
         Shape:
-            If `prepend` is False, the shape of `cdf` will be one smaller than the input `pdf`
-            in all dimensions integrated over.
-            If `prepend` is True, the shape of `cdf` will match that of the input `pdf`.
+        * If `prepend` is False, the shape of `cdf` will be one smaller than the input `pdf`
+        * in all dimensions integrated over.
+        * If `prepend` is True, the shape of `cdf` will match that of the input `pdf`.
 
     """
     # Convert from density to mass using trapezoid rule in each bin
@@ -276,11 +204,29 @@ def cumtrapz(pdf, edges, prepend=True, axis=None):
 
     # Prepend zeros to output array
     if prepend:
-        cdf = _pre_pad_zero(cdf, axis=axis)
+        ndim = _ndim(cdf)
+        temp = [1, 0] if axis is None else [0, 0]
+        padding = [temp for ii in range(ndim)]
+        if axis is not None:
+            padding[axis][0] = 1
+        # cdf = _pre_pad_zero(cdf, axis=axis)
+        cdf = np.pad(cdf, padding, constant_values=0)
 
     return cdf
 
 
+def histogram(data, bins=None, weights=None, density=False, probability=False):
+    if bins is None:
+        bins = 'auto'
+    hist, edges = np.histogram(data, bins=bins, weights=weights, density=False)
+    if density:
+        hist = hist.astype(float) / np.diff(edges)
+    if probability:
+        tot = data.size if (weights is None) else np.sum(weights)
+        hist = hist.astype(float) / tot
+    return hist, edges
+
+
 def matrix_invert(matrix, helper=True):
     try:
         matrix_inv = np.linalg.inv(matrix)
@@ -296,34 +242,46 @@ def meshgrid(*args, indexing='ij', **kwargs):
     return np.meshgrid(*args, indexing=indexing, **kwargs)
 
 
-def midpoints(data, scale='lin', frac=0.5, axis=-1, squeeze=True):
+def midpoints(arr, log=False, frac=0.5, axis=-1, squeeze=False):
     """Return the midpoints between values in the given array.
+
+    If the given array is N-dimensional, midpoints are calculated from the last dimension.
+
+    Arguments
+    ---------
+    arr : ndarray of scalars,
+        Input array.
+    log : bool or None,
+        Find midpoints in log-space
+    frac : float,
+        Fraction of the way between intervals (e.g. `0.5` for half-way midpoints).
+    axis : int, sequence, or `None`,
+        The axis about which to find the midpoints.  If `None`, find the midpoints along all axes.
+        If a sequence (tuple, list, or array), take the midpoints along each specified axis.
+
+    Returns
+    -------
+    mids : ndarray of floats,
+        The midpoints of the input array.
+        The resulting shape will be the same as the input array `arr`, except that
+        `mids.shape[axis] == arr.shape[axis]-1`.
+
     """
 
-    if (np.shape(data)[axis] < 2):
-        raise RuntimeError("Input ``arr`` does not have a valid shape!")
-
-    if scale.lower().startswith('lin'):
-        log = False
-    elif scale.lower().startswith('log'):
-        log = True
+    if axis is None:
+        axis = [ii for ii in range(np.ndim(arr))]
     else:
-        raise ValueError("`scale` must be either 'lin' or 'log'!")
+        axis = np.atleast_1d(axis)
 
     # Convert to log-space
     if log:
-        data = np.log10(data)
+        mids = np.log10(arr)
     else:
-        data = np.array(data)
+        mids = np.array(arr)
 
-    diff = np.diff(data, axis=axis)
-
-    cut = [slice(None)] * data.ndim
-    cut[axis] = slice(0, -1, None)
-    cut = tuple(cut)
-
-    start = data[cut]
-    mids = start + frac*diff
+    # Take the midpoints along each of the desired axes
+    for ax in axis:
+        mids = _midpoints_1d(mids, frac=frac, axis=ax)
 
     if log:
         mids = np.power(10.0, mids)
@@ -333,7 +291,50 @@ def midpoints(data, scale='lin', frac=0.5, axis=-1, squeeze=True):
     return mids
 
 
-def minmax(data, prev=None, stretch=None, log_stretch=None, limit=None):
+def _midpoints_1d(arr, frac=0.5, axis=-1):
+    """Return the midpoints between values in the given array.
+
+    If the given array is N-dimensional, midpoints are calculated from the last dimension.
+
+    Arguments
+    ---------
+    arr : ndarray of scalars,
+        Input array.
+    frac : float,
+        Fraction of the way between intervals (e.g. `0.5` for half-way midpoints).
+    axis : int,
+        Which axis about which to find the midpoints.
+
+    Returns
+    -------
+    mids : ndarray of floats,
+        The midpoints of the input array.
+        The resulting shape will be the same as the input array `arr`, except that
+        `mids.shape[axis] == arr.shape[axis]-1`.
+
+    """
+
+    if not np.isscalar(axis):
+        raise ValueError("Input `axis` argument must be an integer less than ndim={np.ndim(arr)}!")
+
+    if (np.shape(arr)[axis] < 2):
+        raise RuntimeError("Input ``arr`` does not have a valid shape!")
+
+    diff = np.diff(arr, axis=axis)
+
+    # skip the last element, or the last axis
+    # cut = slice_for_axis(arr, axis=axis, stop=-1
+    cut = [slice(None)] * arr.ndim
+    cut[axis] = slice(0, -1, None)
+    cut = tuple(cut)
+
+    start = arr[cut]
+    mids = start + frac*diff
+
+    return mids
+
+
+def minmax(data, positive=False, prev=None, stretch=None, log_stretch=None, limit=None):
     if prev is not None:
         assert len(prev) == 2, "`prev` must have length 2."
     if limit is not None:
@@ -344,7 +345,8 @@ def minmax(data, prev=None, stretch=None, log_stretch=None, limit=None):
         return prev
 
     # Find extrema
-    minmax = np.array([np.min(data), np.max(data)])
+    idx = (data > 0.0) if positive else slice(None)
+    minmax = np.array([np.min(data[idx]), np.max(data[idx])])
 
     # Add stretch (relative to center point)
     if (stretch is not None) or (log_stretch is not None):
@@ -387,21 +389,166 @@ def minmax(data, prev=None, stretch=None, log_stretch=None, limit=None):
     return minmax
 
 
-def modify_exists(path_fname):
-    """Modify the given filename if it already exists.
+def parse_edges(data, edges=None, extrema=None, weights=None, params=None,
+                nmin=5, nmax=1000, pad=None, refine=1.0, bw=None):
     """
-    path_fname = os.path.abspath(path_fname)
-    if not os.path.exists(path_fname):
-        return path_fname
+    """
+    if _ndim(data) not in [1, 2]:
+        err = (
+            "`data` (shape: {}) ".format(np.shape(data))
+            + "must have shape (N,) or (D, N) for `N` data points and `D` parameters!"
+        )
+        raise ValueError(err)
 
-    fname, vers = _fname_match_vers(path_fname)
-    vers = 0 if (vers is None) else vers + 1
-    fname = fname.format(vers)
-    return fname
+    squeeze = (_ndim(data) == 1)
+    data = np.atleast_2d(data)
+    npars = np.shape(data)[0]
+
+    if pad is None:
+        pad = 1 if extrema is None else 0
+
+    extrema = _parse_extrema(data, extrema=extrema, warn=False, params=params)
+
+    # If `edges` provides a specification for each dimension, convert to npars*[edges]
+    if (_ndim(edges) == 0) or (really1d(edges) and (np.size(edges) != npars)):
+        edges = [edges] * npars
+    elif len(edges) != npars:
+        err = "length of `edges` ({}) does not match number of data dimensions ({})!".format(
+            len(edges), npars)
+        raise ValueError(err)
+
+    if bw is None:
+        bw = np.empty((npars, npars), dtype=object)
+    edges = [_get_edges_1d(edges[ii], data[ii], extrema[ii],
+                           npars, nmin, nmax, pad, weights=weights, refine=refine, bw=bw[ii, ii])
+             for ii in range(npars)]
+
+    if squeeze:
+        edges = np.squeeze(edges)
+
+    return edges
 
 
-def percentiles(values, percs=None, sigmas=None, weights=None, axis=None, values_sorted=False):
-    """Compute weighted percentiles.
+def _get_edges_1d(edges, data, extrema, ndim, nmin, nmax, pad, weights=None, refine=1.0, bw=None):
+    """
+
+    Arguments
+    ---------
+    edges : None, int, or array or scalar
+        Specification for bin-edges.
+        `None` : number of bins is automatically calculated from `data`
+        int : used as number of bins, span is calculated from `data`
+        array : used as fixed bin edges, ignores `data`
+    data : 1D array of scalar
+        Data points from which to calculate bin-edges
+    ndim : `None` or int
+        Number of dimensions of the data-set, used to calculate an effective number of data-points.
+
+    """
+
+    if _ndim(edges) == 0:
+        num_bins = edges
+    elif really1d(edges):
+        return edges
+    else:
+        err = "1D `edges` (shape: {}) must be `None`, an integer or a 1D array of edges!".format(
+            np.shape(edges))
+        raise ValueError(err)
+
+    _num_bins, bin_width, _span = _guess_edges(
+        data, extrema=extrema, weights=weights,
+        ndim=ndim, num_min=nmin, num_max=nmax, refine=refine, bw=bw)
+
+    # print("utils.py:_get_edges_1d():")
+    # print("\t", "num_bins = ", num_bins, "_num_bins = ", _num_bins, "refine = ", refine,
+    #       "bin_width = ", bin_width, "_span = ", _span)
+
+    if num_bins is None:
+        num_bins = _num_bins
+
+    if pad is not None:
+        pad_width = pad * bin_width
+        extrema = [extrema[0] - pad_width, extrema[1] + pad_width]
+
+    edges = np.linspace(*extrema, num_bins + 1, endpoint=True)
+    return edges
+
+
+def _guess_edges(data, extrema=None, ndim=None, weights=None, num_min=None, num_max=None, refine=1.0, bw=None):
+    if weights is None:
+        num_eff = data.size
+    else:
+        if (not really1d(weights)) or (np.size(weights) != np.size(data)):
+            err = "Shape of `weights` ({}) does not match `data` ({})!".format(
+                np.shape(weights), np.shape(data))
+            raise ValueError(err)
+
+        num_eff = np.sum(weights)**2 / np.sum(weights**2)
+
+    if (ndim is not None) and (num_eff > 100):
+        # num_eff = np.power(num_eff, 1.0 / ndim)
+        _num_eff = num_eff / ndim**2
+        num_eff = np.clip(_num_eff, 100, None)
+
+    any_inf = [0.0] if not np.iterable(extrema) else [ex for ex in extrema if ex is not None]
+    any_inf = np.any(~np.isfinite(any_inf))
+    if (extrema is None) or any_inf:
+        if any_inf:
+            err = "Given extrema ({}) contain non-finite values!  Overriding!".format(extrema)
+            logging.error(err)
+        idx = np.isfinite(data)
+        extrema = [data[idx].min(), data[idx].max()]
+    span_width = np.diff(extrema)[0]
+
+    # Sturges histogram bin estimator.
+    w1 = span_width / (np.log2(num_eff) + 1.0)
+    # Freedman-Diaconis histogram bin estimator
+    iqr = iqrange(data, log=False, weights=weights)               # get interquartile range
+    w2 = 2.0 * iqr * num_eff ** (-1.0 / 3.0)
+    w3 = bw / np.sqrt(num_eff) if (bw is not None) else 0.0
+
+    widths = [w1, w2, w3]
+    bin_width = [bw for bw in widths if bw > 0.0]
+    if len(bin_width) > 0:
+        bin_width = min(bin_width)
+        bin_width = bin_width / refine
+    else:
+        err = "`bin_width` is not positive (w1 = {}, w2 = {})!".format(w1, w2)
+        logging.warning(err)
+        if np.allclose(data, data[0]):
+            bin_width = 1e-16
+            logging.warning("WARNING: all data is identical! Choosing arbitrary `bin_width`")
+        else:
+            raise ValueError(err)
+
+    if span_width <= 0.0:
+        err = "`span_width` is not positive (span_width={})!".format(span_width)
+        logging.warning(err)
+        if np.allclose(data, data[0]):
+            span_width = 10*bin_width
+            logging.warning("WARNING: all data is identical! Choosing arbitrary `span_width`")
+        else:
+            raise ValueError(err)
+
+    num_bins = int(np.ceil(span_width / bin_width))
+    if (num_min is not None) or (num_max is not None):
+        num_bins = np.clip(num_bins, num_min, num_max)
+
+    # print(f"{num_bins=}, {bin_width=}, {extrema=}, {widths=}")
+
+    return num_bins, bin_width, extrema
+
+
+def iqrange(data, log=False, weights=None):
+    """Calculate inter-quartile range of the given data."""
+    if log:
+        data = np.log10(data)
+    iqr = np.subtract(*quantiles(data, percs=[0.75, 0.25], weights=weights))
+    return iqr
+
+
+def quantiles(values, percs=None, sigmas=None, weights=None, axis=None, values_sorted=False):
+    """Compute weighted quartiles.
 
     Taken from `zcode.math.statistics`
     Based on @Alleo answer: http://stackoverflow.com/a/29677616/230468
@@ -423,11 +570,14 @@ def percentiles(values, percs=None, sigmas=None, weights=None, axis=None, values
         Array of percentiles of the weighted input data.
 
     """
+    if (percs is None) and (sigmas is None):
+        raise ValueError("Either `percs` or `sigmas` must be provided!")
+
     values = np.array(values)
     if percs is None:
         percs = sp.stats.norm.cdf(sigmas)
 
-    if np.ndim(values) > 1:
+    if _ndim(values) > 1:
         if axis is None:
             values = values.flatten()
 
@@ -473,16 +623,67 @@ def really1d(arr):
         Whether `arr` is purely 1D.
 
     """
-    if np.ndim(arr) != 1:
+    if _ndim(arr) != 1:
         return False
     # Empty list or array
     if len(arr) == 0:
         return True
-    if np.any(np.vectorize(np.ndim)(arr)):
+    # Each element must be a scalar
+    if np.any([np.shape(tt) != () for tt in arr]):
+        return False
+
+    return True
+
+
+def flatten(arr):
+    """Flatten a ND array, whether jagged or not, into a 1D array.
+    """
+    # If this is not an iterable, or it's actually 1D then it's already 'flat'
+    if (not np.iterable(arr)) or really1d(arr):
+        return arr
+
+    # Flatten each component and combine recursively
+    return np.concatenate([flatten(aa) for aa in arr])
+
+
+def flatlen(arr):
+    if not np.iterable(arr):
+        return 1
+
+    if really1d(arr):
+        return len(arr)
+
+    return np.sum([flatlen(aa) for aa in arr])
+
+
+def isjagged(arr):
+    """Test if the given array is jagged.
+    """
+    # if np.isscalar(arr) or (np.size(arr) == len(flatten(arr))):
+    if np.isscalar(arr) or (np.size(arr) == flatlen(arr)):
         return False
     return True
 
 
+def jshape(arr, level=0, printout=False, prepend="", indent="  "):
+    """Print the complete shape (even if jagged) of the given array.
+    """
+    arr = np.asarray(arr, dtype=object)
+    if printout:
+        print(prepend + indent*level + str(np.shape(arr)))
+
+    if not isjagged(arr):
+        return np.shape(arr)
+
+    shape = []
+    for aa in arr:
+        sh = jshape(aa, level+1, prepend=prepend, indent=indent, printout=printout)
+        shape.append(sh)
+
+    shape = tuple([np.shape(arr), tuple(shape)])
+    return shape
+
+
 def rem_cov(data, cov=None):
     if cov is None:
         cov = np.cov(*data)
@@ -521,6 +722,25 @@ def spacing(data, scale='log', num=None, dex=10, **kwargs):
     return spaced
 
 
+def stats(data, shape=True, sample=3, stats=True):
+    rv = ""
+    failure = True
+    if shape:
+        failure = False
+        rv += str(np.shape(data))
+    if (sample is not None) and (sample is not False):
+        failure = False
+        rv += " - " + array_str(data, num=sample)
+    if stats:
+        failure = False
+        rv += " - " + stats_str(data)
+
+    if failure:
+        raise ValueError("No stats requested!")
+
+    return rv
+
+
 def stats_str(data, percs=[0.0, 0.16, 0.50, 0.84, 1.00], ave=False, std=False, weights=None,
               format=None, log=False, label_log=True):
     """Return a string with the statistics of the given array.
@@ -583,7 +803,7 @@ def stats_str(data, percs=[0.0, 0.16, 0.50, 0.84, 1.00], ave=False, std=False, w
 
     # Add percentiles
     if percs_flag:
-        tiles = percentiles(data, percs, weights=weights).astype(data.dtype)
+        tiles = quantiles(data, percs, weights=weights).astype(data.dtype)
         out += "(" + ", ".join(form.format(tt) for tt in tiles) + ")"
         out += ", for (" + ", ".join("{:.0f}%".format(100*pp) for pp in percs) + ")"
 
@@ -631,7 +851,7 @@ def trapz_dens_to_mass(pdf, edges, axis=None):
         e.g. `[[x0, x1, ... xn], [y0, y1, ... ym], ...]`
         The length of each sub-list in `edges`, must match the shape of `pdf`.
         e.g. if `edges` is a (3,) list, composed of sub-lists with lengths: `[N, M, L,]` then
-             the shape of `pdf` must be `(N, M, L,)`.
+        the shape of `pdf` must be `(N, M, L,)`.
     axis : int, array_like int, or None
         Along which axes to convert from density to mass.
 
@@ -684,7 +904,7 @@ def trapz_dens_to_mass(pdf, edges, axis=None):
         widths.append(temp)
 
     # Multiply the widths along each dimension to get the volume of each grid cell
-    volumes = np.product(widths, axis=0)
+    volumes = np.product(np.array(widths, dtype=object), axis=0).astype(float)
     # NOTE
     assert np.all(np.shape(volumes) == shp_out), "BAD `volume` shape!"
 
@@ -735,42 +955,6 @@ def run_if_script(func, *args, otherwise=None, **kwargs):
     return run_if(func, target, *args, otherwise=otherwise, **kwargs)
 
 
-def _is_notebook():
-    return _python_environment().startswith('notebook')
-
-
-def _is_script():
-    return _python_environment().startswith('script')
-
-
-def _fname_match_vers(path_fname, digits=2):
-    path, fname = os.path.split(path_fname)
-    match = re.search('_[0-9]{1,}', fname)
-    if match is None:
-        fname_comps = fname.split('.')
-        idx = 0 if (len(fname_comps) == 1) else -2
-        fname_comps[idx] = fname_comps[idx] + '_{{:0{:}d}}'.format(digits)
-        fname = ".".join(fname_comps)
-        fname = os.path.join(path, fname)
-        fname, num = _fname_match_vers(fname.format(0), digits=digits)
-        num = num if os.path.exists(fname.format(0)) else None
-        return fname, num
-
-    match_str = match.group()
-    match_str = match_str.strip('_')
-    num_digits = len(match_str)
-    num = int(match_str)
-
-    form = "_{{:0{:}d}}".format(num_digits)
-    fname = fname.replace('_' + match_str, form)
-    fname = os.path.join(path, fname)
-    num_max = 10**num_digits - 1
-    while os.path.exists(fname.format(num+1)) and (num < num_max - 1):
-        num += 1
-
-    return fname, num
-
-
 def _guess_str_format_from_range(arr, prec=2, log_limit=2, allow_int=True):
     """
     """
@@ -793,16 +977,6 @@ def _guess_str_format_from_range(arr, prec=2, log_limit=2, allow_int=True):
     return form
 
 
-def _pre_pad_zero(aa, axis=None):
-    if axis is None:
-        return np.pad(aa, [1, 0])
-
-    aa = np.moveaxis(aa, axis, 0)
-    aa = np.concatenate([[np.zeros_like(aa[0])], aa], axis=0)
-    aa = np.moveaxis(aa, 0, axis)
-    return aa
-
-
 def _prep_msg(msg=None):
     if (msg is None) or (msg is True):
         msg_fail = "FAILURE:: arrays do not match!"
@@ -831,28 +1005,269 @@ def _python_environment():
         return 'script'
 
 
-class _DummyError(Exception):
-    pass
+def _ndim(vals):
+    import warnings
+    with warnings.catch_warnings():
+        warnings.simplefilter("ignore")
+        return np.ndim(vals)
 
 
-class Test_Base(object):
+def _parse_extrema(data, extrema=None, params=None, warn=True):
+    """Get extrema (min and max) consistent with the given `data`.
 
-    DEF_SEED = 1234
+    `data` must have shape (D, N) for `D` parameters/dimensions, and `N` data points.
 
-    @classmethod
-    def setup_class(cls):
-        np.random.seed(cls.DEF_SEED)
-        return
+    `extrema` can be:
+        None: extrema are calculated
+        (2,): extrema taken as the same for each dimension
+        (D, 2): extrema given for each dimension
 
-    # Override `__getattribute__` to print the class and function name whenever they are called
-    def __getattribute__(self, attr):
-        value = object.__getattribute__(self, attr)
-        if attr.startswith('__'):
-            return value
+        In the latter two cases, any values of `extrema` that are 'None' are filled in using
+        extrema from the data.
 
-        name = object.__getattribute__(self, "__class__").__name__
-        if callable(value):
-            print("\n|{}:{}|".format(name, attr))
+    """
 
-        np.random.seed(object.__getattribute__(self, "DEF_SEED"))
-        return value
+    # `data` must be shaped as (D, N)
+    if _ndim(data) != 2:
+        err = "`data` must be shaped (D, N) for `D` dimensions/parameters and `N` data points!"
+        raise ValueError(err)
+
+    npars = np.shape(data)[0]
+
+    data_extrema = [minmax(dd) for dd in data]
+    if extrema is not None:
+        extrema = copy.deepcopy(extrema)
+
+    if extrema is None:
+        extrema = data_extrema
+    elif (params is not None) and (len(extrema) != npars):
+        extrema = [extrema[pp] for pp in params]
+
+    # Check components of given `extrema` to make sure they are valid
+    #   fill in any `None` values with extrema from the data
+    else:
+        # Convert from (2,) ==> (D, 2)
+        if really1d(extrema) and (np.shape(extrema) == (2,)):
+            extrema = [extrema for ii in range(npars)]
+        # If already (D, 2) we're good, keep going
+        elif np.shape(np.array(extrema, dtype=object)) == (npars, 2):
+            pass
+        # If jagged (D,) array
+        elif len(extrema) == npars:
+            if really1d(extrema) or not np.all([(ee is None) or (len(ee) == 2) for ee in extrema]):
+                err = "Each element of jagged `extrema` must be `None` or have length 2!"
+                raise ValueError(err)
+        # Otherwise bad
+        else:
+            err = "`extrema` shape '{}' unrecognized for {} parameters!".format(
+                jshape(extrema), npars)
+            raise ValueError(err)
+
+        # Fill in `None` values and check if given `extrema` is out of bounds for data
+        for dd in range(npars):
+            if extrema[dd] is None:
+                extrema[dd] = [None, None]
+
+            for ii in range(2):
+                if extrema[dd][ii] is None:
+                    extrema[dd][ii] = data_extrema[dd][ii]
+
+            if warn:
+                if (extrema[dd][0] > data_extrema[dd][0]):
+                    msg = "lower `extrema` in dimension {} ({}) is above data min: {}!".format(
+                        dd, extrema[dd][0], data_extrema[dd][0])
+                    logging.warning(msg)
+
+                if (extrema[dd][1] < data_extrema[dd][1]):
+                    msg = "lower `extrema` in dimension {} ({}) is below data max: {}!".format(
+                        dd, extrema[dd][1], data_extrema[dd][1])
+                    logging.warning(msg)
+
+    return extrema
+
+
+def _random_data_1d_01(num=1e4):
+    num = int(num)
+    np.random.seed(12345)
+    _d1 = np.random.normal(4.0, 1.0, num//2)
+    _d2 = np.random.lognormal(0, 0.5, size=num - _d1.size)
+    data = np.concatenate([_d1, _d2])
+
+    xx = np.linspace(0.0, 7.0, 200)[1:]
+    yy = 0.5*np.exp(-(xx - 4.0)**2/2) / np.sqrt(2*np.pi)
+    yy += 0.5 * np.exp(-np.log(xx)**2/(2*0.5**2)) / (0.5*xx*np.sqrt(2*np.pi))
+
+    truth = [xx, yy]
+    return data, truth
+
+
+def _random_data_1d_02(num=1e4):
+    num = int(num)
+    np.random.seed(12345)
+    _d1 = np.random.normal(1.0, 1.0, num//2)
+    _d2 = np.random.uniform(1.0, 3.0, size=num - _d1.size)
+    data = np.concatenate([_d1, _d2])
+
+    # xx = np.linspace(0.0, 7.0, 200)[1:]
+    # yy = 0.5*np.exp(-(xx - 4.0)**2/2) / np.sqrt(2*np.pi)
+    # yy += 0.5 * np.exp(-np.log(xx)**2/(2*0.5**2)) / (0.5*xx*np.sqrt(2*np.pi))
+    #
+    # truth = [xx, yy]
+    return data
+
+
+def _random_data_2d_01(num=1e3, noise=0.2):
+    num = int(num)
+
+    sigma = [1.0, 0.2]
+    corr = [
+        [+1.5, -0.5],
+        [-0.5, +1.0]
+    ]
+
+    cov = np.zeros_like(corr)
+    for (ii, jj), cc in np.ndenumerate(corr):
+        cov[ii, jj] = cc * sigma[ii] * sigma[jj]
+
+    data = np.random.multivariate_normal(np.zeros_like(sigma), cov, num).T
+    dd = data[1, :]
+    dd = (dd - dd.min())/dd.max()
+    data *= np.sqrt(dd)[np.newaxis, :]
+
+    nn = int(num*noise)
+    noise = [np.random.normal(data[ii, :].mean(), 2*ss, nn) for ii, ss in enumerate(sigma)]
+    sel = np.random.choice(num, nn, replace=False)
+    data[:, sel] = np.array(noise)
+    return data
+
+
+def _random_data_2d_02(num=1e3, noise=0.2):
+    num = int(num)
+
+    xx = np.random.normal(1.0, 0.5, num)
+    yy = np.random.uniform(0.0, 2.0, num)
+    data = [xx, yy]
+    data = [np.sort(aa) for aa in data]
+    xx, yy = data
+    idx = np.arange(num)
+    sel = np.random.choice(num, num//4, replace=False)
+    idx[sel] = np.random.choice(num, num//4, replace=False)
+    # np.random.shuffle(yy[np.random.choice(num, num//2, replace=False)])
+    data = [xx, yy[idx]]
+
+    # idx = np.arange(num)
+    # np.random.shuffle(idx)
+    # data = [zz[idx] for zz in data]
+    # idx = np.random.choice(num, num//2, replace=False)
+    # np.random.shuffle(np.array(data)[:, idx])
+
+    return data
+
+
+def _random_data_2d_03(num=1e3):
+    num = int(num)
+
+    aa = np.random.lognormal(0.0, 0.3, size=3*num) - 1
+    aa += np.random.uniform(-1.0, 1.0, aa.size)
+    aa = aa[aa > 0.0]
+    aa = np.random.choice(aa, size=num, replace=False)
+    cc = np.random.power(2, size=num)
+
+    # Make `aa` and `cc` strongly covariant
+    COV = 0.1    # the smaller the value, the stronger the covariance
+    aa = np.sort(aa)
+    c1 = cc / cc.max()
+    a1 = (aa / aa.max())**(1/4)
+    xx = a1**2 + c1**2 + np.random.normal(0.0, COV, size=a1.size)
+    idx = np.argsort(xx)
+    cc = cc[idx]
+
+    #    unsort `aa` and keep covariance
+    idx = np.arange(aa.size)
+    np.random.shuffle(idx)
+    aa = aa[idx]
+    cc = cc[idx]
+
+    data = [aa, cc]
+    return data
+
+
+def _random_data_3d_01(num=1e3):
+    num = int(num)
+
+    sigma = [1.0, 0.2, 1.5]
+    corr = [
+        [+1.4, +0.8, +0.4],
+        [+0.8, +1.0, -0.25],
+        [+0.4, -0.25, +1.0]
+    ]
+
+    cov = np.zeros_like(corr)
+    for (ii, jj), cc in np.ndenumerate(corr):
+        cov[ii, jj] = cc * sigma[ii] * sigma[jj]
+
+    data = np.random.multivariate_normal(np.zeros_like(sigma), cov, num).T
+    dd = data[1, :]
+    dd = (dd - dd.min())/dd.max()
+    data *= np.sqrt(dd)[np.newaxis, :]
+
+    pc = 0
+    extr = [np.percentile(dd, [0+pc, 100-pc]) for dd in data]
+    noise = [np.random.uniform(*ex, num//5) for ex in extr]
+    data = np.append(data, noise, axis=1)
+
+    return data
+
+
+def _random_data_3d_02(num=1e3, noise=0.2):
+    num = int(num)
+
+    sigma = [1.0, 0.2, 1.5]
+    corr = [
+        [+1.4, +0.8, +0.4],
+        [+0.8, +1.0, -0.25],
+        [+0.4, -0.25, +1.0]
+    ]
+
+    cov = np.zeros_like(corr)
+    for (ii, jj), cc in np.ndenumerate(corr):
+        cov[ii, jj] = cc * sigma[ii] * sigma[jj]
+
+    data = np.random.multivariate_normal(np.zeros_like(sigma), cov, num).T
+    dd = data[1, :]
+    dd = (dd - dd.min())/dd.max()
+    data *= np.sqrt(dd)[np.newaxis, :]
+
+    nn = int(num*noise)
+    noise = [np.random.normal(data[ii, :].mean(), 2*ss, nn) for ii, ss in enumerate(sigma)]
+    sel = np.random.choice(num, nn, replace=False)
+    data[:, sel] = np.array(noise)
+    return data
+
+
+def _random_data_3d_03(num=1e3, par=[0.0, 0.5], cov=0.1):
+    num = int(num)
+
+    aa = np.random.lognormal(*par, size=3*num) - 1
+    aa = aa[aa > 0.0]
+    aa = np.random.choice(aa, size=num, replace=False)
+    bb = np.random.normal(scale=par[1], size=num)
+    cc = np.random.power(2, size=num)
+
+    # Make `aa` and `cc` strongly covariant
+    # cov = 0.1    # the smaller the value, the stronger the covariance
+    aa = np.sort(aa)
+    c1 = cc / cc.max()
+    a1 = (aa / aa.max())**(1/4)
+    xx = a1**2 + c1**2 + np.random.normal(0.0, cov, size=a1.size)
+    idx = np.argsort(xx)
+    cc = cc[idx]
+
+    #    unsort `aa` and keep covariance
+    idx = np.arange(aa.size)
+    np.random.shuffle(idx)
+    aa = aa[idx]
+    cc = cc[idx]
+
+    data = [aa, bb, cc]
+    return data
